/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 */

#include <asterisk/file.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/options.h>
#include <asterisk/../../astconf.h>
#include <asterisk/pbx.h>
#include <asterisk/module.h>
#include <asterisk/lock.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

static char *synopsis = "Use Sox to stream a file to a channel";
static char *app = "Sox";
static char *tdesc = "Use Sox to stream a file to a channel";

static char *sox = "/usr/bin/sox";
static char *usage = "Sox requires arguments (filename[|options])\n";

static char *descrip = "Usage: Sox(<filename>[|options])\n"
"\nOptions:\n\n"
"n[oanswer]   -- Do not auto-answer the channel.\n"
"s[kip]       -- Skip if the channel is not up.\n\n";



STANDARD_LOCAL_USER;

LOCAL_USER_DECL;

#define BUF_SIZE 320
#define ALL_DONE(u,ret) LOCAL_USER_REMOVE(u); return ret;

struct genobj {
	int fd;
	int rf;
	int wf;
	int running;
	struct ast_frame fr;
	char *filename;
};

static void *sox_alloc(struct ast_channel *chan, void *params) {
	struct genobj *obj = params;
	int pid = 0;
	int fds[2];
	
	obj->rf=chan->readformat;
	obj->wf=chan->writeformat;
	if (ast_set_read_format(chan,AST_FORMAT_SLINEAR) < 0) {
        ast_log(LOG_ERROR,"Could Not Set Read Format.\n");
		return NULL;
    }
	if (ast_set_write_format(chan,AST_FORMAT_SLINEAR) < 0) {
		ast_log(LOG_ERROR,"Could Not Set Write Format.\n");
		return NULL;
	}
	
	obj->fr.frametype = AST_FRAME_VOICE;
    obj->fr.subclass = AST_FORMAT_SLINEAR;
    obj->fr.offset = AST_FRIENDLY_OFFSET;
    obj->fr.mallocd = 0;
	obj->running = 1;
	if (pipe(fds)) {
        ast_log(LOG_WARNING, "Pipe failed\n");
		return NULL;
    }
	pid = fork();
	if(pid < 0) {
		if(fds[0])
			close(fds[0]);
		if(fds[1])
			close(fds[1]);
		return NULL;
	}
	if(!pid) {
		int x;
        close(fds[0]);
        /* Stdout goes to pipe */
        dup2(fds[1], STDOUT_FILENO);
        /* Close unused file descriptors */
        for (x=3;x<8192;x++)
            close(x);

		execl(sox,sox,"-c","1",obj->filename,"-t","raw","-r","8000","-c","1","-s","-w","-",NULL);
		ast_log(LOG_WARNING,"DoH! execl Failed!\n");
        if(fds[1])
            close(fds[1]);
		exit(0);
	}
	obj->fd = fds[0];
	return obj;
}

static void sox_release(struct ast_channel *chan, void *params) {
	struct genobj *obj = params;
	if(obj->fd) {
		close(obj->fd);
	}

	if (ast_set_read_format(chan,obj->rf) < 0) {
        ast_log(LOG_ERROR,"Could Not Set Read Format.\n");
    }
	if (ast_set_write_format(chan,obj->wf) < 0) {
		ast_log(LOG_ERROR,"Could Not Set Write Format.\n");
	}	

}

static int sox_generate(struct ast_channel *chan, void *data, int len, int samples) {
	struct genobj *obj = data;
	int res = 0;
	char buf[BUF_SIZE];
	struct pollfd fds[1];

	if(!(obj && obj->fd)) {
		obj->running = -1;
		return -1;
	}
	

	fds[0].fd = obj->fd;
	fds[0].events = POLLIN;
	res = poll(fds, 1, 150);
	if (res < 0) {
		if (errno != EINTR)
			ast_log(LOG_WARNING, "poll returned < 0: %s\n", strerror(errno));
		obj->running = 0;
		return -1;
	}
	if(!res) {
		obj->running = 0;
		return -1;
	}
	if (fds[0].revents) {
		if ((res = read(obj->fd,buf,BUF_SIZE)) < 1) {
			if (res)
				ast_log(LOG_WARNING, "Short read (%d) (%s)!\n", res, strerror(errno));
			obj->running = 0;
			return -1;
		}
		obj->fr.samples = res/2;
		obj->fr.datalen = res;
		obj->fr.data = buf;
		ast_write(chan,&obj->fr);
	} else {
		obj->running = 0;
		return -1;
	}
	if(res != BUF_SIZE) {
		obj->running = 0;
		return -1;
	}
	
	return 0;
}

static struct ast_generator soxgen = 
{
	alloc: sox_alloc,
	release: sox_release,
	generate: sox_generate,
};



static int sox_exec(struct ast_channel *chan, void *data)
{
	struct localuser *u;
	struct genobj obj;
	char *options=NULL, *filename=NULL,*buf=NULL;
	int res=0;
	size_t size;

	if (!data) {
		ast_log(LOG_WARNING, usage);
		return -1;
	}
	filename = ast_strdupa((char *) data);
	if((options = strchr(filename,'|'))) {
		*options = '\0';
		options++;
		if((strchr(options,'s') || !strcmp(options,"skip")) && chan->_state != AST_STATE_UP)
			return 0;
		
		if(!strchr(options,'n') || strcmp(options,"noanswer")) {
			ast_answer(chan);
		}
	}


	LOCAL_USER_ADD(u);
	memset(&obj,0,sizeof(struct genobj));

	if (filename && filename[0] != '/') {
		size = strlen(filename) + strlen((char *)ast_config_AST_VAR_DIR) + 20;
		buf = alloca(size);
		snprintf(buf, size, "%s/%s/%s", (char *)ast_config_AST_VAR_DIR, "sounds", filename);
		filename = buf;
	}

	obj.filename = filename;

	if (option_verbose > 2)
		ast_verbose(VERBOSE_PREFIX_3 "Sox Playing '%s'\n", filename);
	ast_activate_generator(chan, &soxgen, &obj);

	while(obj.running == 1) {
		res = ast_waitfordigit(chan, 100);
		if(res < 0 || res == '*')
			break;
	}
	
	ast_deactivate_generator(chan);
	LOCAL_USER_REMOVE(u);
	return obj.running;
}

int unload_module(void)
{
	STANDARD_HANGUP_LOCALUSERS;
	return ast_unregister_application(app);
}

int load_module(void)
{
	return ast_register_application(app, sox_exec, synopsis, descrip);
}

char *description(void)
{
	return tdesc;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
